import numpy as np
from scipy.io import readsav
import astropy.units as u
from .mag_field_wrapper import MagFieldWrapper
from .mag_field_lin_fff import MagFieldLinFFF

__author__     = "Alexey G. Stupishin"
__email__      = "agstup@yandex.ru"
__copyright__  = "SUNCAST project, 2024"
__license__    = "MIT"
__version__    = "1.1.3"
__maintainer__ = "Alexey G. Stupishin"
__status__     = "beta"

# import pydoc

def mfp_util_invert_index(index, sizes):
    kx = index % sizes[0]
    kyz = index // sizes[0]
    ky = kyz % sizes[1]
    kz = kyz // sizes[1]
    
    return (kz*sizes[0] + kx)*sizes[1] + ky

def mfp_util_invert_index_array(data, sizes):
    _data = data.copy()
    for k in range(0, len(data)):
        _data[k] = mfp_util_invert_index(data[k], sizes)

    return _data

def mfp_util_transpose_index(data, sizes, transpose):
    if not transpose:
        return data
    
    _data = data.copy()
    for k in range(0, len(data)):
        inv = mfp_util_invert_index(k, sizes)
        _data[inv] = data[k]
        
    return _data

class MagFieldProcessor(MagFieldWrapper):
    FIELD_NONE =  0
    FIELD_BOUND = 1
    FIELD_NLFFF = 2

    #-------------------------------------------------------------------------------
    @staticmethod
    def _as_dict(record_):
        return {name:record_[name] for name in record_.dtype.names}

    #-------------------------------------------------------------------------------
    def __init__(self, lib_path = ""):
        super().__init__(lib_path)
        self.__bx    = None
        self.__by    = None
        self.__bz    = None
        self.__step  = None
        self.__status  = self.FIELD_NONE
        self.__weight_bound_size = 0.1

    #-------------------------------------------------------------------------------
    def energy(self, box = None, weight_bound_size = None, dr = None):
        if box is None:
            box = self.get_field_cube()
        if weight_bound_size is None:
            weight_bound_size = self.__weight_bound_size
        if dr is None:
            dr = self.__step

        assert dr is not None
            
        dr = dr.to(u.cm).value
        if np.isscalar(dr):
            dr = [dr, dr, dr]
        
        N = np.array(np.array(box['bx']).shape)
        left = np.floor(weight_bound_size*N).astype(np.int32)
        right = np.ceil((1.0-weight_bound_size)*N).astype(np.int32)
        absB2 = (box['bx'][left[0]:right[0],left[1]:right[1],0:right[2]]**2 
               + box['by'][left[0]:right[0],left[1]:right[1],0:right[2]]**2 
               + box['bz'][left[0]:right[0],left[1]:right[1],0:right[2]]**2 
                )
        totalB2 = np.sum(absB2)
        volume = np.prod(dr)

        return totalB2 / 8 / np.pi * volume

    #-------------------------------------------------------------------------------
    def load_cube_vars(self, box, dr = 0.001*u.solRad):
        """
            Set initial magnetic field components.

                Parameters
                ----------
                box['bx'], by, bz : float

                dr : array-like or singleton astropy.Quantity
                    
                Returns
                -------
                box : dict
                    A dictionary with keys "bx", "by", "bz"
                    bx, by, bz : ndarray
                        3D np.float64 arrays (input copy)
                
        """

        swap = (2,0,1)
        return self.__load_vars(box['by'].transpose(swap), box['bx'].transpose(swap), box['bz'].transpose(swap), dr)

    #-------------------------------------------------------------------------------
    def load_bottom(self, bottom):
        assert self.__bx is not None and self.__by is not None and self.__bz is not None
        
        cube = self.get_field_cube()
        cube['bx'][:,:,0] = bottom['by'].transpose(1,0).astype(np.float64, order="C")
        cube['by'][:,:,0] = bottom['bx'].transpose(1,0).astype(np.float64, order="C")
        cube['bz'][:,:,0] = bottom['bz'].transpose(1,0).astype(np.float64, order="C")
        self.__status  = self.FIELD_BOUND
        
        swap = (2,1,0)
        return self.__load_vars(cube['bx'].transpose(swap), cube['by'].transpose(swap), cube['bz'].transpose(swap), self.__step)

    #-------------------------------------------------------------------------------
    def LFFF_bounded(self, bottom, pad = (1, 1), nz = None, dr = 0.001*u.solRad, alpha = 0):
        
        lfff = MagFieldLinFFF.create_LFFF_cube(bottom['bz'].astype(np.float64, order="C"), pad = pad, nz = nz, alpha = alpha)
        self.load_cube_vars(lfff, dr = dr)

        return self.load_bottom(bottom)

    #-------------------------------------------------------------------------------
    def sav_to_cube(self, filename):

        sav_data = readsav(filename, python_dict = True)

        box = sav_data.get('box', sav_data.get('pbox'))

        return self._as_dict(box[0])
        
    #-------------------------------------------------------------------------------
    def load_cube_sav(self, filename):

        box = self.sav_to_cube(filename)
        
        swap = (0,2,1)
        return self.__load_vars(np.transpose(box['BY'], swap), np.transpose(box['BX'], swap), np.transpose(box['BZ'], swap), box['DR'] * u.solRad)
    
    #-------------------------------------------------------------------------------
    def __load_vars(self, bx, by, bz, dr):
        self.__bx = bx.astype(np.float64, order="C")
        self.__by = by.astype(np.float64, order="C")
        self.__bz = bz.astype(np.float64, order="C")

        if np.isscalar(dr):
            self.__step = [dr, dr, dr]
        else:
            self.__step = dr
            
        self.status = self.FIELD_BOUND

        return self.get_field_cube()

    #-------------------------------------------------------------------------------
    def get_field_cube(self, rc = 0):
        assert self.__bx is not None and self.__by is not None and self.__bz is not None
        swap = (1,2,0)
        return dict(bx = self.__by.transpose(swap).copy(), by = self.__bx.transpose(swap).copy(), bz = self.__bz.transpose(swap).copy(), rc = rc)

    #-------------------------------------------------------------------------------
    def NLFFF(self, **kwargs):
        """
            Wrapper to external call of Weighted Wiegelmann NLFF Field Reconstruction Method.
            Magnetic field cube should be preliminary set. Field components modified "in place".

                Parameters (**kwargs)
                ---------------------
                weight_bound_size : float, optional
                    Bounary buffer zone size (in parts of corresponding dim. size), 
                    default is 0.1 (i.e. buffer zone is 10% of dimension size from all boundaries, 
                    except photosphere plain). 
                    Use weight_bound_size = 0 for no-buffer-zone approach.
                    default = 0.1

                dense_grid_use : bool, optional
                    Use condensed grid approach. `True` is recommended.
                    default = True
                
                derivative_stencil : int, optional
                    Number of stencil points for derivatives
                    (experimental, internal use).
                    default = 3

                Returns
                -------
                box : dict
                    A dictionary with keys "bx", "by", "bz", "rc"
                    bx, by, bz : ndarray
                        3D np.float64 arrays of magnetic field components
                    rc : int
                        zero if OK, otherwise reconstruction error (to be specified)
                
        """
        
        assert self.__bx is not None and self.__by is not None and self.__bz is not None

        rc = super().NLFFF_wrapper(self.__bx, self.__by, self.__bz, **kwargs)

        self.__weight_bound_size = super().get_double('weight_bound_size')

        return self.get_field_cube(rc = rc)

#-------------------------------------------------------------------------------
    def lines(self
            , reduce_passed = None
            , chromo_level = 1
            , seeds = None
            , max_length = 0
            , step = 1.0
            , tolerance = 1e-3
            , tolerance_bound = 1e-3
            , n_processes = 0
            , debug_input = False
             ):
        """
            max_length == 0: no coordinates will be stored
            max_length > 0: max. number of coordinates to store
            max_length < 0: max. number of coordinates to store will be automatically estimated
        """
        
        assert self.__bx is not None and self.__by is not None and self.__bz is not None

        res = super().lines_wrapper(self.__bx, self.__by, self.__bz, reduce_passed, chromo_level, seeds, max_length, step, tolerance, tolerance_bound, n_processes, debug_input)
        
        print(res)

        N = np.flip(self.__bx.shape)
        transpose = seeds is None
        res['voxel_status'] = mfp_util_transpose_index(res['voxel_status'], N, transpose)
        res['phys_length'] = mfp_util_transpose_index(res['phys_length'], N, transpose)
        res['av_field'] = mfp_util_transpose_index(res['av_field'], N, transpose)
        res['codes'] = mfp_util_transpose_index(res['codes'], N, transpose)

        res['apex_idx'] = mfp_util_transpose_index(mfp_util_invert_index_array(res['apex_idx'], N), N, transpose)
        res['start_idx'] = mfp_util_transpose_index(mfp_util_invert_index_array(res['start_idx'], N), N, transpose)
        res['end_idx'] = mfp_util_transpose_index(mfp_util_invert_index_array(res['end_idx'], N), N, transpose)
        res['seed_idx'] = mfp_util_transpose_index(mfp_util_invert_index_array(res['seed_idx'], N), N, transpose)
        
        if res['coords'] is not None:
            res['coords'] = res['coords'][:, [1,0,2,3]]

        return res        
 
# pydoc.writedoc("MagFieldProcessor")



import numpy as np
import warnings

__author__     = "Alexey G. Stupishin"
__email__      = "agstup@yandex.ru"
__copyright__  = "SUNCAST project, 2024-2025"
__credits__    = ["Sergey A. Anfinogentov"]
__license__    = "MIT"
__version__    = "1.1.3"
__maintainer__ = "Alexey G. Stupishin"
__status__     = "beta"

class MagFieldLinFFF:
    FIELD_NONE = 0    
    FIELD_BOTTOM = 1  
    FIELD_LFFF = 2  

    #-------------------------------------------------------------------------------
    def __init__(self, bottom = None, pad = (1, 1)):
        self.__status = self.FIELD_NONE
        if bottom is None:
            self.__field_fft = None
        else:
            self.set_field(bottom, pad)
            
        pass

    #-------------------------------------------------------------------------------
    @staticmethod
    def create_LFFF_cube(bz, pad = (1, 1), nz = None, alpha = 0, directive_cosines = (0, 0, 1)):
        lin_proc = MagFieldLinFFF()
        return lin_proc.LFFF_cube(bottom = bz, pad = pad, nz = nz, alpha = alpha, directive_cosines = directive_cosines)    
        
    #-------------------------------------------------------------------------------
    def LFFF_cube(self, bottom = None, pad = (1, 1), nz = None, alpha = 0, directive_cosines = (0, 0, 1)):
        
        if bottom is not None:
            self.set_field(bottom, pad)

        assert self.__field_fft is not None, 'No bottom field set'

        if nz is None:
            nz = np.ceil(np.max(self.__size)*0.7).astype(np.int32)

        bx = np.zeros((self.__size[0], self.__size[1], nz))
        by = np.zeros((self.__size[0], self.__size[1], nz))        
        bz = np.zeros((self.__size[0], self.__size[1], nz))        

        for k in range(0, nz):
            res = self.lfff_at_z(k, alpha, directive_cosines)
            bx[:,:,k] = res['bx']
            by[:,:,k] = res['by']
            bz[:,:,k] = res['bz']
            pass

        self.__status = self.FIELD_LFFF

        return dict(bx = bx
                  , by = by
                  , bz = bz
                   )

    #-------------------------------------------------------------------------------
    def set_field(self, bottom, pad = (1, 1)):
        # prepare
        size = np.shape(bottom)
        pad_size = np.ceil(size * (1 + np.array(pad, dtype = np.float64, order = 'C'))).astype(int)
        pad_size_half = (pad_size+1) // 2
        pad_size = pad_size_half * 2
        v = np.sum(bottom)
        s0 = np.prod(size)
        sp = np.prod(pad_size)
        d_pad = v/(sp-s0)
        
        field_pad = np.zeros(pad_size, dtype = np.float64, order = 'C') - d_pad
        field_pad[:size[0], :size[1]] = bottom

        # self.__field_av = v/s0;
    
        # FFT2
        self.__field_fft = np.fft.fft2(field_pad)
    
        # uv-domain coefficients
        u_vect = np.concatenate((np.linspace(0, pad_size_half[0], num = pad_size_half[0], endpoint = False), -np.linspace(pad_size_half[0], 0, num = pad_size_half[0], endpoint = False))) / pad_size[0]
        self.__u = np.tile(u_vect, (pad_size[1], 1)).transpose()
        v_vect = np.concatenate((np.linspace(0, pad_size_half[1], num = pad_size_half[1], endpoint = False), -np.linspace(pad_size_half[1], 0, num = pad_size_half[1], endpoint = False))) / pad_size[1]
        self.__v = np.tile(v_vect, (pad_size[0], 1))

        self.__q = np.sqrt(self.__u**2 + self.__v**2)

        self.__size = size
        self.__status = self.FIELD_BOTTOM

        pass

    #-------------------------------------------------------------------------------
    def lfff_at_z(self, z, alpha = 0, directive_cosines = (0, 0, 1)):
        
        if alpha == 0:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                invqv = 1 / (self.__u*directive_cosines[0] + self.__v*directive_cosines[1] + 1j*self.__q*directive_cosines[2])
            invqv[np.isinf(invqv) | np.isnan(invqv)] = 0

            G = self.__field_fft * invqv * np.exp(-2*np.pi*self.__q*z)

            bx = (np.fft.ifft2(self.__u * G)).real
            by = (np.fft.ifft2(self.__v * G)).real
            bz = (np.fft.ifft2(1j*self.__q * G)).real # + self.__field_av / directive_cosines[2]
            pass
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                k = np.sqrt(4*(np.pi*self.__q)**2 - alpha**2);
            k[k.imag != 0 | np.isinf(k) | np.isnan(k)] = 0
    
            ukva = self.__u*k - self.__v*alpha
            vkua = self.__v*k + self.__u*alpha
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                invqv = 1 / (-1j*(directive_cosines[0]*ukva + directive_cosines[1]*vkua)/(2*np.pi*self.__q**2) + directive_cosines[2])
            invqv[np.isinf(invqv) | np.isnan(invqv)] = 0

            G = self.__field_fft * invqv * np.exp(-k*z)
            
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                indq = 1 / (2*np.pi*self.__q**2);
            indq[np.isinf(indq) | np.isnan(indq)] = 0
    
            bx = (np.fft.ifft2(-1j*ukva*indq * G)).real
            by = (np.fft.ifft2(-1j*vkua*indq * G)).real
            bz = np.fft.ifft2(G).real # + self.__field_av / directive_cosines[2]
            pass

        return dict(bx = bx[:self.__size[0], :self.__size[1]]
                  , by = by[:self.__size[0], :self.__size[1]]
                  , bz = bz[:self.__size[0], :self.__size[1]]
                   )



#!/usr/bin/env python3

import platform
from setuptools import Extension, setup
from pathlib import Path
import pdb
import sys

link_flags = {
    'Linux': ["-shared", "-lm", "-pthread"],
    'Windows': ["-shared", "-lm", "-pthread"],
    'Darwin': ["-lm", "-pthread"]
}

current_os = platform.system()

if current_os in link_flags:
    extra_link = link_flags[current_os]
else:
    extra_link = []

rp = Path("./AMaFiL")

header_dirs = ["./" + str(rp / p) for p in ("Common/BaseMath",    "Common/BasePhysics",
                                     "Common/General",     "Common/IDL",
                                     "Common/Multithread", "Common/Utils", "sources")]

source_dirs = [rp / p for p in ("Common/BaseMath",
                                "Common/BasePhysics",
                                "Common/Multithread",
                                "Common/Utils", "sources")]

source_files = []
for d in source_dirs:
    source_files.extend([str(d / x.name) for x in sorted(d.glob("*.cpp"))])
#pdb.set_trace()
source_files.append("pyinit.cpp")

setup(
    ext_modules=[
        Extension(
            name="pyAMaFiL.WWNLFFFReconstruction",
            sources = source_files,
            include_dirs = header_dirs,
            export_symbols = ["utilInitialize", "utilSetInt", "utilSetDouble", "utilGetVersion", "mfoNLFFFCore", "mfoGetLines"],
            language = "c++",
            extra_compile_args = ["-std=c++11", "-fPIC", "-O2", "-fpermissive"],
            extra_link_args = extra_link
        ),
    ]
)